# Helper for querying the dbus

package require dbus
package require tdom

proc qdbus {bus args} {
    global dbusresult
    set dbus [dbus connect $bus]
    set argc [llength $args]
    set args [lassign $args name path method]
    try {
	if {$argc == 0} {
	    # Get all bus names
	    set svc [dbus info service]
	    set path [dbus info path]
	    set owner {}
	    foreach n [dbuscall $dbus -dest $svc $path $svc ListNames] {
		set o [dbuscall $dbus -dest $svc $path $svc GetNameOwner $n]
		if {$o eq $n} {
		    dict lappend owner $o
		} else {
		    dict lappend owner $o $n
		}
	    }
	    return $owner
	}
	if {$argc == 1} {
	    # Get paths
	    return [dbustree $dbus $name]
	}
	set svc [dbus info service].Introspectable
	set xml [dbuscall $dbus -dest $name $path $svc Introspect]
	dom parse $xml doc
	if {$argc == 2} {
	    foreach n [$doc selectNodes /node/interface] {
		set intf [$n getAttribute name]
		set prop {}
		set slot {}
		foreach s [$n childNodes] {
		    # Skip comments
		    if {[$s nodeType] ne "ELEMENT_NODE"} continue
		    if {[$s nodeName] eq "property"} {
			dict set prop [$s getAttribute name ""] $s
		    } else {
			dict set slot [$s getAttribute name ""] $s
		    }
		}
		# properties
		dict for {key p} $prop {
		    set access [$p getAttribute access readwrite]
		    set t [$p selectNodes {
			./annotation[@name="com.trolltech.QtDBus.QtTypeName"]
		    }]
		    if {[llength $t]} {
			set type [$t getAttribute value ""]
		    } else {
			set type [$p getAttribute type ""]
		    }
		    dict set rc $intf property $key [list $access $type]
		}
		# methods and signals
		dict for {key s} $slot {
		    set str [$s nodeName]
		    set out ""
		    set args {}
		    foreach a [$s selectNodes ./arg] {
			set dir [$a getAttribute direction in]
			if {$dir eq "out"} {
			    set out [$a getAttribute type ""]
			} else {
			    set sig [$a getAttribute type ""]
			    if {[$a hasAttribute name]} {
				lappend args [$a getAttribute name]:$sig
			    } else {
				lappend args $sig
			    }
			}
		    }
		    dict set rc $intf $str $key \
		      [dict create args $args return $out]
		}
	    }
	    return $rc
	}
	if {![regexp {(.*)\.(.*)} $method -> intf member]} {
	    set intf $name
	    set member $method
	}
	foreach node [$doc selectNodes {/node/interface[@name=$intf] \
	  /*[@name=$member and (name()="method" or name()="property")]}] {
	    set type [$node nodeName]
	    if {$type eq "method"} {
		set sig [lmap n [$node selectNodes {arg[@direction="in"]}] {
		    $n getAttribute type @
		}]
		if {[llength $sig] == [llength $args]} {
		    return [dbuscall $dbus -dest $name \
		      -signature [join $sig ""] $path $intf $member {*}$args]
		}
	    } else {
		if {[llength $args] == 0} {
		    return [dbuscall $dbus -dest $name -signature ss \
		      $path org.freedesktop.DBus.Properties Get \
		      $intf $member]
		} elseif {[llength $args] == 1} {
		    set sig [$node getAttribute type @]
		    return [dbuscall $dbus -dest $name -signature ssv \
		      $path org.freedesktop.DBus.Properties Set \
		      $intf $member [list $sig [lindex $args 0]]]
		}
	    }
	}
	return -code error "method $member not known in interface $intf"
    } finally {
	dbus close $dbus
    }
}

proc dbuscall {dbus args} {
    global dbusresult
    set lambda [list apply [list {info args} {
	global dbusresult
	if {[dict get $info messagetype] eq "error"} {
	    set code [split [string toupper [dict get $info errorname]] .]
	    set dbusresult [list -code error -errorcode $code [lindex $args 0]]
	} else {
	    set dbusresult [lrange $args 0 0]
	}
    }]]
    dbus call $dbus -handler $lambda {*}$args
    vwait dbusresult
    return {*}$dbusresult
}

proc dbustree {dbus name {path /}} {
    set rc $path
    set pfx [if {$path ne "/"} {set path}]
    set svc [dbus info service].Introspectable
    set xml [dbuscall $dbus -dest $name $path $svc Introspect]
    dom parse $xml doc
    foreach n [$doc selectNodes /node/node] {
	set node [$n getAttribute name]
	lappend rc {*}[dbustree $dbus $name $pfx/$node]
    }
    return $rc
}
